#!/usr/bin/env python
#
# BioLite - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012-2013 Brown University. All rights reserved.
# 
# This file is part of BioLite.
# 
# BioLite is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# BioLite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with BioLite.  If not, see <http://www.gnu.org/licenses/>.

import os
import re

from Bio import SeqIO

from biolite import config
from biolite import diagnostics
from biolite import utils
from biolite import wrappers2


trinity_header_pattern = re.compile(\
	r'c(\d+).graph_c(\d+)_seq(\d+) len=(\d+) path=\[([-: \d]+)\]$')

standard_contig_header = \
	">Locus_{0}_Transcript_{1}/0_Confidence_{2}_Length_{3}"


def write_trinity(components, assembly):
	with open(assembly, 'w') as f:
		for component in components:
			fasta = component+'.graph.allProbPaths.fasta'
			if os.path.exists(fasta):
				for record in SeqIO.parse(fasta, 'fasta'):
					match = trinity_header_pattern.match(record.description)
					if not match is None:
						comp, sub, seq, length, path = match.groups()
						print >>f, standard_contig_header.format(
							"%s.%s" % (comp, sub), seq,
							sum(c == ' ' for c in path), length)
						print >>f, record.seq
					else:
						utils.info(
							"unparsable header for component",
							os.path.basename(component))
			else:
				utils.info(
					"no output for component",
					os.path.basename(component))


def quantify(assembly, name, reads, outdir, workdir=os.getcwd()):
	"""
	"""

	prefix = os.path.join(workdir, name)
	wrappers2.RsemReference(assembly, prefix)

	# Could do our own mapping in the future:
	# bowtie -p 2 -n 2 -l 25 -e 99999999 -m 200 ref -1 input1 -2 input2 -S out

	# Bowtie2 - RSEM doesn't support gapped alignments
	#   (https://groups.google.com/forum/#!topic/rsem-users/l9vzo4HGSeQ)
	# "--all --ignore-quals -N 1"

	name = os.path.join(outdir, name)
	wrappers2.RsemExpression(reads, prefix, name)

	genes = name + '.genes.results'
	isoforms = name + '.isoforms.results'

	diagnostics.log_path(genes, 'genes')
	diagnostics.log_path(isoforms, 'isoforms')

	return genes


def fpkms(coverage_table):
	"""
	"""

	fpkms = {}

	with open(coverage_table) as f:
		next(f) # burn the header line
		for line in f:
			fields = line.rstrip().split()
			assert len(fields) == 7
			fpkms[fields[0]] = float(fields[6])

	return fpkms

